<?php
// $Id: apc.inc,v 1.1.2.1 2009/07/20 09:17:39 doq Exp $

/**
 * APC cache engine.
 */
class apcCacheEngine extends CacheEngine {
  /**
   * @return
   *   TRUE if cache engine is available; FALSE otherwise.
   */
  function status() {
    return function_exists('apc_fetch');
  }

  /**
   * @return
   *   TRUE if page_fast_cache enabled.
   */
  function page_fast_cache() {
    return $this->fast_cache;
  }
  
  /**
   * Return item from cache if it is available.
   *
   * @param string $key
   *   The key to fetch.
   * @return mixed object|bool
   *   Returns either the cache object or FALSE on failure
   */
  function get($key) {
    $key = $this->key($key);

    $cache = parent::get($key);
    if (isset($cache)) {
      return $cache;
    }

    $cache = apc_fetch($key);
    if (is_object($cache) && $cache->serialized) {
      $cache->data = unserialize($cache->data);
    }
    parent::set($key, $cache);
    return $cache;
  }

  /**
   * Add item into cache.
   *
   * @param string $key
   *   The key to set.
   * @param string $value
   *   The data to store in the cache.
   * @param string $expire
   *   The time to expire in seconds.
   * @param string $headers
   *   The page headers.
   * @return bool
   *   Returns TRUE on success or FALSE on failure
   */
  function set($key, $value, $expire = CACHE_PERMANENT, $headers = NULL) {
    $key = $this->key($key);

    $time = time();
    if ($expire == CACHE_TEMPORARY) {
      $expire = $time + 180;
    }
    // Create new cache object.
    $cache = new stdClass;
    $cache->cid = $key;
    $cache->created = $time;
    $cache->expire = $expire;
    $cache->headers = $headers;

    if ($expire != CACHE_PERMANENT) {
      // Convert Drupal $expire, which is a timestamp, to a TTL for APC.
      $expire -= $time;
    }
    if (!is_string($value)) {
      $cache->serialized = TRUE;
      $cache->data = serialize($value);
    }
    else { 
      $cache->serialized = FALSE;
      $cache->data = $value;
    }

    $return = FALSE;
    if (!empty($key) && $this->lock()) {
      // Get lookup table to be able to keep track of bins.
      $lookup = apc_fetch($this->lookup);

      // If the lookup table is empty, initialize table.
      if (empty($lookup)) {
        $lookup = array();
      }

      // Set key to 1 so we can keep track of the bin.
      $lookup[$key] = 1;

      // Attempt to store full key and value.
      if (!apc_store($key, $cache, $expire)) {
        unset($lookup[$key]);
        $return = FALSE;
      }
      else {
        // Update static cache.
        parent::set($key, $cache);
        $return = TRUE;
      }
      
      // Resave the lookup table (even on failure).
      apc_store($this->lookup, $lookup);
      
      // Remove lock.
      $this->unlock();
    }

    return $return;
  }

  /**
   * Remove item from cache.
   *
   * @param string $key
   *   The key to set.
   * @return mixed object|bool
   *   Returns either the cache object or FALSE on failure
   */
  function delete($key) {
    $key = $this->key($key);

    // Remove from static array cache.
    parent::delete($key);

    if (substr($key, -1, 1) == '*') {
      $key = substr($key, 0, strlen($key) - 1);
      $lookup = apc_fetch($this->lookup);
      if (!is_array($lookup)) {
        $lookup = array();
      }
      foreach ($lookup as $k => $v) {
        if (substr($k, 0, strlen($key) - 1)) {
          apc_delete($k);
          unset($lookup[$k]);  
        }
      }
      if ($this->lock()) {
        apc_store($this->lookup, $lookup);
        $this->unlock();
      }
    }
    else {
      if (!empty($key)) {
        if (!apc_delete($key)) {
          $return = FALSE;
        }
        else {
          $return = TRUE;
        }
      }
    }
  }

  /**
   * Flush the entire cache.
   *
   * @param none
   * @return mixed bool
   *   Returns TRUE
   */
  function flush() {
    $key = $this->key($key);

    parent::flush();
    if ($this->lock()) {
      // Get lookup table to be able to keep track of bins.
      $lookup = apc_fetch($this->lookup);

      // If the lookup table is empty, remove lock and return.
      if (empty($lookup)) {
        $this->unlock();
        return TRUE;
      }

      // Cycle through keys and remove each entry from the cache.
      foreach ($lookup as $k => $v) {
        if (apc_delete($k)) {
          unset($lookup[$k]);
        }
      }

      // Resave the lookup table (even on failure).
      apc_store($this->lookup, $lookup);

      // Remove lock.
      $this->unlock();
    }

    return TRUE;
  }

  /**
   * Lock the cache from other writes.
   *
   * @param none
   * @return string
   *   Returns TRUE on success, FALSE on failure
   */
  function lock() {
    static $new_apc;
    if (!isset($new_apc)) {
      $new_apc = function_exists('apc_add');
    }
    
    if ($new_apc) {
      // Lock once by trying to add lock file, if we can't get the lock, we will loop
      // for 3 seconds attempting to get lock. If we still can't get it at that point,
      // then we give up and return FALSE.
      if (apc_add($this->lock, TRUE) === FALSE) {
        $time = time();
        while (apc_add($this->lock, TRUE) === FALSE) {
          if (time() - $time >= 3) {
            return FALSE;
          }
        }
      }
    }
    else {
      // For older versions of APC (before 3.0.13).
      $time = time();
      while (apc_fetch($this->lock)) {
        if (time() - $time >= 3) {
          return FALSE;
        }
      }
      apc_store($this->lock, TRUE);
    }

    return TRUE;
  }

  /**
   * Lock the cache from other writes.
   *
   * @param none
   * @return bool
   *   Returns TRUE on success, FALSE on failure
   */
  function unlock() {
    return apc_delete($this->lock);
  }

  /**
   * Statistics information.
   */
  function stats() {
    $apc_stats = apc_cache_info('user', TRUE);
    $apc_stats['uptime'] = time() - $apc_stats['start_time'];
    $stats = array(
      'uptime' => $apc_stats['uptime'],
      'bytes_used' => $apc_stats['mem_size'],
      'bytes_total' => ini_get('apc.shm_size') * 1024 * 1024,
      'gets' => $apc_stats['num_hits'],
      'sets' => $apc_stats['num_inserts'],
      'hits' => $apc_stats['num_hits'] + $apc_stats['num_misses'],
      'misses' => $apc_stats['num_misses'],
      'req_rate' => (($apc_stats['num_hits'] + $apc_stats['num_misses']) / $apc_stats['uptime']),
      'hit_rate' => ($apc_stats['num_hits'] / $apc_stats['uptime']),
      'miss_rate' => ($apc_stats['num_misses'] / $apc_stats['uptime']),
      'set_rate' => ($apc_stats['num_inserts'] / $apc_stats['uptime']),
    );
    return $stats;
  }

}
